/* 
 * $Id: send_admin.c,v 1.1.2.1 1998/05/07 06:15:29 sumikawa Exp $
 */

/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
 * Copyright(C)1997 by Hitachi, Ltd.
 * Hitachi Id: send_admin.c,v 1.2 1997/12/22 09:56:49 sumikawa Exp $
 */

/* header include files */
#include "defs.h"

/* private variables declarations */
static struct prefix pref_val;

/* 
 * Func: send_admin_table 
 * Desc: to send routing table information to admin.
 * INs : struct prefix *send_pref 
 * ret : none 
 * Call: get_start_ptr(), send_all_entries().
 */
void
send_admin_table(struct prefix *send_pref)
{
	struct tree_node *start_ptr;
	char buffer[ADM_PKTSIZE];
	int buf_used;

	if (!rnhead)
		goto SEND_EOF;	/* now flushing */
	bzero(&pref_val, sizeof(pref_val));
	pref_val = *send_pref;
/* 
 * If complete table is requested, start_ptr is head of tree.
 */
	if (pref_val.prf_len == 0) {
		start_ptr = &(rnhead->th_node[TREE_HEAD]);
	} else {
		if ((start_ptr = get_start_ptr()) == NULL)
			goto SEND_EOF;
	}

	bzero(buffer, ADM_PKTSIZE);
	buf_used = 0;
	send_all_entries(start_ptr, buffer, &buf_used);

      SEND_EOF:
	bzero(buffer, ADM_PKTSIZE);
	buf_used = 1;
	buffer[0] = ADM_EOF;
	if (sendto(admin_sock, buffer, buf_used, 0, (struct sockaddr *)&admin_dest,
		   sizeof(admin_dest)) < 0) {
		syslog(LOG_ERR, "sendto failed in send_admin_table : %m");
		return;
	}
	return;
}				/* end of send_admin_table() */

/* 
 * Func: get_start_ptr 
 * Desc: to get the start entry from where all entries are to be sent. 
 * INs : none 
 * ret : struct tree_node *start_here 
 * Call: 
 */
struct tree_node *
get_start_ptr(void)
{
	struct tree_node *start_here;
	struct route_entry rte;

	bzero((char *)&rte, sizeof(rte));
	rte.rip6_addr = pref_val.prf_addr;
	rte.rip6_prflen = pref_val.prf_len;

	start_here = &(rnhead->th_node[TREE_HEAD]);
	while ((start_here->tn_bposn != LEAF_BIT_POSN)
	       && (start_here->tn_bposn < pref_val.prf_len)) {
		if (pref_val.prf_addr.s6_addr[start_here->boff] & start_here->tn_bmask) {
			start_here = start_here->rptr;
		} else {
			start_here = start_here->lptr;
		}
	}

	return (start_here);
}

#define PRINT_NODE( next_ptr ) \
do\
{\
  struct rt_plen   *plen_ptr;\
  struct rt_table  *tbl_ptr;\
\
  if( !prefcmp( &(next_ptr->key), &(pref_val.prf_addr), pref_val.prf_len ))\
    break;\
  for( plen_ptr = next_ptr->dst; plen_ptr; plen_ptr = plen_ptr->rp_next )\
  {\
    if( plen_ptr->rp_len < pref_val.prf_len )continue;\
    if( (ADM_PKTSIZE - *buf_used) < sizeof( struct rt_table ) )\
    {\
      if( sendto( admin_sock, buffer, *buf_used, 0,\
                  (struct sockaddr *)&admin_dest, sizeof(admin_dest))<0)\
      {\
        syslog(LOG_ERR, "sendto failed in send_admin_table : %m");\
      }\
      bzero( buffer, sizeof(ADM_PKTSIZE) );\
      *buf_used = 0;\
    }\
    tbl_ptr = (struct rt_table *)( buffer + *buf_used );\
    tbl_ptr->rt_dest   = next_ptr->key;\
    tbl_ptr->rt_prflen = plen_ptr->rp_len;\
    tbl_ptr->rt_metric = plen_ptr->rp_metric;\
    tbl_ptr->rt_flag   = plen_ptr->rp_state;\
    tbl_ptr->rt_gway   = plen_ptr->rp_gway->gw_addr;\
    strcpy( tbl_ptr->rt_ifname, plen_ptr->rp_gway->gw_ifp->if_name );\
    *buf_used += sizeof(struct rt_table);\
  }\
}while(0)

/* 
 * Func: send_all_entries
 * Desc: to send entry
 * INs : struct tree_node *start_ptr, char *buffer, int *buf_used
 * ret : none
 * Call: none
 */
void
send_all_entries(struct tree_node *start_ptr, char *buffer, int *buf_used)
{
	struct tree_node *next_ptr;

	/* step 0: start */
	next_ptr = start_ptr;
	if (next_ptr->tn_bposn == LEAF_BIT_POSN) {
		PRINT_NODE(next_ptr);
		goto LASTONE;
	}
	/* step 1: left side of start  */
	/* ASSERT: start is not a LEAF */
	while (1) {
	/* step 1: find leftmost leaf */
		while (next_ptr->tn_bposn != LEAF_BIT_POSN) {
			next_ptr = next_ptr->lptr;
		}

		PRINT_NODE(next_ptr);

	/* step 2: go back to unprocessed parent */
	/* ASSERT: next_ptr -> LEAF              */
		while (next_ptr->tn_backp->rptr == next_ptr) {
			next_ptr = next_ptr->tn_backp;
			if (next_ptr == start_ptr)
				goto LASTONE;
		}

	/* step 3: one step down right */
		next_ptr = next_ptr->tn_backp->rptr;
	}

 LASTONE:
	if (*buf_used) {
		if (sendto(admin_sock, buffer, *buf_used, 0,
		  (struct sockaddr *)&admin_dest, sizeof(admin_dest)) < 0) {
			syslog(LOG_ERR, "sendto failed in send_admin_table : %m");
		}
	}
	return;
}
