/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *		PPP IP Protocol Interface
 *
 *	    Written by Toshiharu OHNO (tony-o@iij.ad.jp)
 *
 *   Copyright (C) 1993, Internet Initiative Japan, Inc. All rights reserverd.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Internet Initiative Japan.  The name of the
 * IIJ may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: ip.c,v 1.1.1.2.2.1 1998/04/20 13:02:45 shin Exp $
 *
 *	TODO:
 *		o Return ICMP message for filterd packet
 *		  and optionaly record it into log.
 */

#ifdef INET

#include <sys/param.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_tun.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#ifndef NOALIAS
#include <alias.h>
#endif
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "command.h"
#include "mbuf.h"
#include "log.h"
#include "id.h"
#include "defs.h"
#include "timer.h"
#include "fsm.h"
#include "lcpproto.h"
#include "hdlc.h"
#include "loadalias.h"
#include "vars.h"
#include "arp.h"
#include "filter.h"
#include "os.h"
#include "vjcomp.h"
#include "lcp.h"
#include "modem.h"
#include "tun.h"
#include "multiproto.h"
#include "ipcp.h"
#include "ip.h"

extern void SendPppFrameInet(struct mbuf *bp);
extern void LcpClose();

int ipInOctets, ipOutOctets;
int ipConnectSecs, ipIdleSecs;

static struct mqueue IpOutputQueues[PRI_FAST + 1];
static struct in_aliasreq ifra;

static int
IpIsAddrUnspec(struct sockaddr *sa)
{
  return (((struct sockaddr_in *)sa)->sin_addr.s_addr == INADDR_ANY);
}

static void
IpMakeAddrUnspec(struct sockaddr *sa)
{
  ((struct sockaddr_in *)sa)->sin_addr.s_addr = INADDR_ANY;
}

static int
IpLoopbackCheck(void *pr_hdr_p)
{
  return (((struct ip *) pr_hdr_p)->ip_dst.s_addr == IpcpInfo.want_ipaddr.s_addr) ? 1 : 0;
}

static void
IpInitSockMaskFromLen(struct sockaddr *sa, int prefixlen)
{
  ((struct sockaddr_in *)sa)->sin_addr.s_addr = netmasks[prefixlen];
  if (prefixlen) {
    sa->sa_len = afsw[PPP_AFINET].AfAddrOffInSock + (prefixlen >> 3);
    if (prefixlen & 0x7)
      sa->sa_len++;
  }
}

static void
IpStartOutput()
{
  struct mqueue *queue;
  struct mbuf *bp;
  int cnt;

  if (IpcpFsm.state != ST_OPENED)
    return;
  for (queue = &IpOutputQueues[PRI_FAST]; queue >= IpOutputQueues; queue--) {
    if (queue->top) {
      bp = Dequeue(queue);
      if (bp) {
	cnt = plength(bp);
	SendPppFrameInet(bp);
	RestartIdleTimer(AF_INET);
	IpcpAddOutOctets(cnt);
	break;
      }
    }
  }
}

void
IcmpError(void *pr_hdr_p, int code)
{
#ifdef notdef
  struct mbuf *bp;

  if (pip->ip_p != IPPROTO_ICMP) {
    bp = mballoc(cnt, MB_IPIN);
    memcpy(MBUF_CTOP(bp), ptr, cnt);
    SendPppFrameInet(bp);
    RestartIdleTimer(AF_INET);
    IpcpAddOutOctets(cnt);
  }
#endif
}

/*
 *  IP related initialization
 */
void
IpInit(void)
{
  af_sock2ppp[AF_INET] = PPP_AFINET;
  af_ppp2sock[PPP_AFINET] = AF_INET;
  afsw[PPP_AFINET].AfAddrOffInSock = 4; /* off to sin_addr */
  afsw[PPP_AFINET].AfMyAddrLabel = "MYADDRINET";
  afsw[PPP_AFINET].AfFlags = PPP_AFFLAGS_HAS_ALIAS|PPP_AFFLAGS_SUPPORTED;
  afsw[PPP_AFINET].AfIdleTime = &ipIdleSecs;
  afsw[PPP_AFINET].AfLogLevel = LogIPCP;
  afsw[PPP_AFINET].AfMbInType = MB_IPIN;
  afsw[PPP_AFINET].AfMbQType = MB_IPQ;
  afsw[PPP_AFINET].AfFsm = &IpcpFsm;
  afsw[PPP_AFINET].AfOutputQueues = IpOutputQueues;
  afsw[PPP_AFINET].AfMyAddr = &IpcpInfo.want_ipaddr;
  afsw[PPP_AFINET].AfHisAddr = &IpcpInfo.his_ipaddr;
  afsw[PPP_AFINET].AfErrorReport = IcmpError;
  afsw[PPP_AFINET].AfInitSockMaskFromLen = IpInitSockMaskFromLen;
  afsw[PPP_AFINET].AfInput = IpInput;
  afsw[PPP_AFINET].AfMakeAddrUnspec = IpMakeAddrUnspec;
  afsw[PPP_AFINET].AfStartOutput = IpStartOutput;
  afsw[PPP_AFINET].AfSendPppFrame = SendPppFrameInet;
  afsw[PPP_AFINET].AfPacketLen = IpPacketLen;
  afsw[PPP_AFINET].AfIsAddrUnspec = IpIsAddrUnspec;
  afsw[PPP_AFINET].AfLoopbackCheck = IpLoopbackCheck;
  afsw[PPP_AFINET].AfSetAddress = SetIpaddress;
  afsw[PPP_AFINET].AfResetAddress = ResetIpaddress;
}

u_short
IpPacketLen(void *pr_hdr_p)
{
  return ntohs(((struct ip *)pr_hdr_p)->ip_len);
}

void
IpInput(struct mbuf * bp)
{				/* IN: Pointer to IP pakcet */
  u_char *cp;
  struct mbuf *wp;
  int nb, nw;
  struct tun_data tun;

  tun_fill_header(tun, AF_INET);
  cp = tun.udata.data;
  nb = 0;
  for (wp = bp; wp; wp = wp->next) {	/* Copy to contiguous region */
    if (sizeof tun.udata.data - (cp - tun.udata.data) < wp->cnt) {
      LogPrintf(LogERROR, "IpInput: Packet too large (%d) - dropped\n",
                plength(bp));
      pfree(bp);
      return;
    }
    memcpy(cp, MBUF_CTOP(wp), wp->cnt);
    cp += wp->cnt;
    nb += wp->cnt;
  }

#ifndef NOALIAS
  if (mode & MODE_ALIAS) {
    struct tun_data *frag;
    int iresult;
    char *fptr;

    iresult = VarPacketAliasIn(tun.udata.data, sizeof tun.udata.data);
    nb = ntohs(((struct ip *) tun.udata.data)->ip_len);

    if (nb > MAX_MRU) {
      LogPrintf(LogERROR, "IpInput: Problem with IP header length\n");
      pfree(bp);
      return;
    }
    if (iresult == PKT_ALIAS_OK
	|| iresult == PKT_ALIAS_FOUND_HEADER_FRAGMENT) {
      if (PacketCheck(AF_INET, tun.udata.data, nb, FL_IN) < 0) {
	pfree(bp);
	return;
      }
      IpcpAddInOctets(nb);

      nb = ntohs(((struct ip *) tun.udata.data)->ip_len);
      nb += sizeof tun - sizeof tun.udata.data;
      nw = write(tun_out, &tun, nb);
      if (nw != nb)
        if (nw == -1)
	  LogPrintf(LogERROR, "IpInput: wrote %d, got %s\n", nb,
                    strerror(errno));
        else
	  LogPrintf(LogERROR, "IpInput: wrote %d, got %d\n", nb, nw);

      if (iresult == PKT_ALIAS_FOUND_HEADER_FRAGMENT) {
	while ((fptr = VarPacketAliasGetFragment(tun.udata.data)) != NULL) {
	  VarPacketAliasFragmentIn(tun.udata.data, fptr);
	  nb = ntohs(((struct ip *) fptr)->ip_len);
          frag = (struct tun_data *)
	    ((char *)fptr - sizeof tun + sizeof tun.udata.data);
          nb += sizeof tun - sizeof tun.udata.data;
	  nw = write(tun_out, frag, nb);
	  if (nw != nb)
            if (nw == -1)
	      LogPrintf(LogERROR, "IpInput: wrote %d, got %s\n", nb,
                        strerror(errno));
            else
	      LogPrintf(LogERROR, "IpInput: wrote %d, got %d\n", nb, nw);
	  free(frag);
	}
      }
    } else if (iresult == PKT_ALIAS_UNRESOLVED_FRAGMENT) {
      nb = ntohs(((struct ip *) tun.udata.data)->ip_len);
      nb += sizeof tun - sizeof tun.udata.data;
      frag = (struct tun_data *)malloc(nb);
      if (frag == NULL)
	LogPrintf(LogALERT, "IpInput: Cannot allocate memory for fragment\n");
      else {
        tun_fill_header(*frag, AF_INET);
	memcpy(frag->udata.data,
	       tun.udata.data, nb - sizeof tun + sizeof tun.udata.data);
	VarPacketAliasSaveFragment(frag->udata.data);
      }
    }
  } else
#endif /* #ifndef NOALIAS */
  {			/* no aliasing */
    if (PacketCheck(AF_INET, tun.udata.data, nb, FL_IN) < 0) {	    
      pfree(bp);
      return;
    }
    IpcpAddInOctets(nb);
    nb += sizeof tun - sizeof tun.udata.data;
    nw = write(tun_out, &tun, nb);
    if (nw != nb)
      if (nw == -1)
	LogPrintf(LogERROR, "IpInput: wrote %d, got %s\n", nb, strerror(errno));
      else
        LogPrintf(LogERROR, "IpInput: wrote %d, got %d\n", nb, nw);
  }
  pfree(bp);

  RestartIdleTimer(AF_INET);
}

#if 0
int
IsIpEnqueued()
{
  struct mqueue *queue;
  int exist = 0;

  for (queue = &IpOutputQueues[PRI_FAST]; queue >= IpOutputQueues; queue--) {
    if (queue->qlen > 0) {
      exist = 1;
      break;
    }
  }
  return (exist);
}
#endif

static struct in_addr oldmine, oldhis;

static int
SetIpDevice(struct in_addr myaddr,
	    struct in_addr hisaddr,
	    struct in_addr netmask,
	    enum set_method how)	    
{
  struct sockaddr_in *sin;
  int s;
  u_long mask, addr;
  extern struct ifreq ifrq;  

  s = ID0socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    LogPrintf(LogERROR, "SetIpDevice: socket(): %s\n", strerror(errno));
    return (-1);
  }
  /* Init ifra's ifname from ifrq */
  strncpy(ifra.ifra_name, ifrq.ifr_name, IFNAMSIZ - 1);
  if (how == SET_DOWN) {
    if (Enabled(ConfProxy))
      cifproxyarp(s, oldhis);	
    if (oldmine.s_addr == 0 && oldhis.s_addr == 0) {
      close(s);
      return (0);
    }
    memset(&ifra.ifra_addr, '\0', sizeof ifra.ifra_addr);
    memset(&ifra.ifra_broadaddr, '\0', sizeof ifra.ifra_broadaddr);
    memset(&ifra.ifra_mask, '\0', sizeof ifra.ifra_mask);    
    if (ID0ioctl(s, SIOCDIFADDR, &ifra) < 0) {
      LogPrintf(LogERROR, "SetIpDevice: ioctl(SIOCDIFADDR): %s\n",
		strerror(errno));
      close(s);
      return (-1);
    }
    oldmine.s_addr = oldhis.s_addr = 0;
  } else {
    /*
     * If given addresses are alreay set, then ignore this request.
     */
    if (oldmine.s_addr == myaddr.s_addr && oldhis.s_addr == hisaddr.s_addr) {
      close(s);
      return (0);
    }

    /*
     * If different address has been set, then delete it first.
     */
    if (oldmine.s_addr || oldhis.s_addr) {
      memset(&ifra.ifra_addr, '\0', sizeof ifra.ifra_addr);
      memset(&ifra.ifra_broadaddr, '\0', sizeof ifra.ifra_broadaddr);
      memset(&ifra.ifra_mask, '\0', sizeof ifra.ifra_mask);
      if (ID0ioctl(s, SIOCDIFADDR, &ifra) < 0) {
        LogPrintf(LogERROR, "SetIpDevice: ioctl(SIOCDIFADDR): %s\n",
		  strerror(errno));
        close(s);
        return (-1);
      }
    }

    /*
     * Set interface address
     */
    sin = (struct sockaddr_in *) & (ifra.ifra_addr);
    sin->sin_family = AF_INET;
    sin->sin_addr = myaddr;
    sin->sin_len = sizeof(*sin);

    /*
     * Set destination address
     */
    sin = (struct sockaddr_in *) & (ifra.ifra_broadaddr);
    sin->sin_family = AF_INET;
    sin->sin_addr = hisaddr;
    sin->sin_len = sizeof(*sin);

    /*
     * */
    addr = ntohl(myaddr.s_addr);
    if (IN_CLASSA(addr))
      mask = IN_CLASSA_NET;
    else if (IN_CLASSB(addr))
      mask = IN_CLASSB_NET;
    else
      mask = IN_CLASSC_NET;

    /*
     * if subnet mask is given, use it instead of class mask.
     */
    if (netmask.s_addr && (ntohl(netmask.s_addr) & mask) == mask)
      mask = ntohl(netmask.s_addr);

    sin = (struct sockaddr_in *) & (ifra.ifra_mask);
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = htonl(mask);
    sin->sin_len = sizeof(*sin);

    if (ID0ioctl(s, SIOCAIFADDR, &ifra) < 0) {
      if (how != SET_TRY)
        LogPrintf(LogERROR, "SetIpDevice: ioctl(SIOCAIFADDR): %s\n",
		  strerror(errno));
      close(s);
      return (-1);
    }
    oldhis.s_addr = hisaddr.s_addr;
    oldmine.s_addr = myaddr.s_addr;
    if (Enabled(ConfProxy))
      sifproxyarp(s, hisaddr);
  }
  close(s);
  return (0);
}

int
TrySetIpaddress(void *myaddr, void *hisaddr)
{
  return (SetIpDevice(*(struct in_addr *)myaddr,
		      *(struct in_addr *)hisaddr, ifnetmask, SET_TRY));
}

int
SetIpaddress(void * myaddr, void * hisaddr)
{
  return (SetIpDevice(*(struct in_addr *)myaddr, *(struct in_addr *)hisaddr, 
		      ifnetmask, SET_UP));
}

int
ResetIpaddress(void)
{
  struct in_addr zeroaddr;
  
  zeroaddr.s_addr = 0;
  return (SetIpDevice(zeroaddr, zeroaddr, zeroaddr, SET_DOWN));
}

#endif /* INET */
