/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *		PPP IPV6 Protocol Interface
 *
 *         created by copying and modifing ip.c
 *
 * $Id: ipv6.c,v 1.1.4.3 1998/06/29 10:52:41 shin Exp $
 *
 */

/*
 *		PPP IP Protocol Interface
 *
 *	    Written by Toshiharu OHNO (tony-o@iij.ad.jp)
 *
 *   Copyright (C) 1993, Internet Initiative Japan, Inc. All rights reserverd.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Internet Initiative Japan.  The name of the
 * IIJ may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: ipv6.c,v 1.1.4.3 1998/06/29 10:52:41 shin Exp $
 *
 *	TODO:
 *		o Return ICMP message for filterd packet
 *		  and optionaly record it into log.
 */

#ifdef INET6
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifndef NOALIAS
#include <alias.h>
#endif /* NOALIAS */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "defs.h"
#include "timer.h"
#include "fsm.h"
#include "lcpproto.h"
#include "command.h"
#include "hdlc.h"
#include "loadalias.h"
#include "filter.h"
#include "mbuf.h"
#include "log.h"
#include "id.h"
#include "os.h"
#include "tun.h"
#include "multiproto.h"
#include "ipv6cp.h"
#include "ipv6.h"

extern void SendPppFrameInet6(struct mbuf *bp);
extern void LcpClose();

int ipv6InOctets, ipv6OutOctets;
int ipv6ConnectSecs, ipv6IdleSecs;
u_char ll_prefixlen = 64; /* default for IPv6 link local address */

static struct ppp_in6_ifaddr *ppp_in6_ifaddr = 0;
static struct mqueue Ipv6OutputQueues[PRI_FAST + 1];
static struct in6_aliasreq ifra;

static int
Ipv6IsAddrUnspec(struct sockaddr *sa)
{
  return (IN6_IS_ADDR_UNSPECIFIED(&((struct sockaddr_in6 *)sa)->sin6_addr));
}

static void
Ipv6MakeAddrUnspec(struct sockaddr *sa)
{
  ((struct sockaddr_in6 *)sa)->sin6_addr = in6addr_any;
}

static int
Ipv6LoopbackCheck(void *pr_hdr_p)
{
  return (IN6_ARE_ADDR_EQUAL(&((struct ip6_hdr *)pr_hdr_p)->ip6_dst, &Ipv6MyAddress)) ? 1 : 0;
}

static void
Ipv6InitSockMaskFromLen(struct sockaddr *sa, int prefixlen)
{
  ((struct sockaddr_in6 *)sa)->sin6_addr = NetmasksInet6[prefixlen];
  if (prefixlen) {
    sa->sa_len = afsw[PPP_AFINET6].AfAddrOffInSock + (prefixlen >> 3);
    if (prefixlen & 0x7)
      sa->sa_len++;
  }
}

static void
Icmpv6Error(void *pr_hdr_p, int code)
{
  /* notyet */
}

static u_short
Ipv6PacketLen(void *pr_hdr_p)
{
  return ntohs(((struct ip6_hdr *)pr_hdr_p)->ip6_plen);
}

static void
Ipv6StartOutput()
{
  struct mqueue *queue;
  struct mbuf *bp;
  int cnt;

  if (Ipv6cpFsm.state != ST_OPENED)
    return;
  for (queue = &Ipv6OutputQueues[PRI_FAST]; queue >= Ipv6OutputQueues; queue--) {
    if (queue->top) {
      bp = Dequeue(queue);
      if (bp) {
	cnt = plength(bp);
	SendPppFrameInet6(bp);
	RestartIdleTimer(AF_INET6);
	Ipv6cpAddOutOctets(cnt);
	break;
      }
    }
  }
}

/*
 *  IPV6 related initialization
 */
void
Ipv6Init(void)
{
  af_sock2ppp[AF_INET6] = PPP_AFINET6;
  af_ppp2sock[PPP_AFINET6] = AF_INET6;
  afsw[PPP_AFINET6].AfAddrOffInSock = /* off to sin6_addr */
#ifdef SIN6_IFINDEX
    16;
#else /* SIN6_IFINDEX */
    8;
#endif /* SIN6_IFINDEX */
  afsw[PPP_AFINET6].AfMyAddrLabel = "MYADDRINET6";
  afsw[PPP_AFINET6].AfFlags = PPP_AFFLAGS_SUPPORTED;
  afsw[PPP_AFINET6].AfIdleTime = &ipv6IdleSecs;
  afsw[PPP_AFINET6].AfLogLevel = LogIPV6CP;
  afsw[PPP_AFINET6].AfMbInType = MB_IPV6IN;
  afsw[PPP_AFINET6].AfMbQType = MB_IPV6Q;
  afsw[PPP_AFINET6].AfFsm = &Ipv6cpFsm;
  afsw[PPP_AFINET6].AfOutputQueues = Ipv6OutputQueues;
  afsw[PPP_AFINET6].AfMyAddr = &Ipv6MyAddress;
  afsw[PPP_AFINET6].AfHisAddr = &Ipv6HisAddress;
  afsw[PPP_AFINET6].AfErrorReport = Icmpv6Error;
  afsw[PPP_AFINET6].AfInitSockMaskFromLen = Ipv6InitSockMaskFromLen;
  afsw[PPP_AFINET6].AfInput = Ipv6Input;
  afsw[PPP_AFINET6].AfMakeAddrUnspec = Ipv6MakeAddrUnspec;
  afsw[PPP_AFINET6].AfStartOutput = Ipv6StartOutput;
  afsw[PPP_AFINET6].AfSendPppFrame = SendPppFrameInet6;  
  afsw[PPP_AFINET6].AfPacketLen = Ipv6PacketLen;
  afsw[PPP_AFINET6].AfIsAddrUnspec = Ipv6IsAddrUnspec;
  afsw[PPP_AFINET6].AfLoopbackCheck = Ipv6LoopbackCheck;
  afsw[PPP_AFINET6].AfSetAddress = SetIpv6address;
  afsw[PPP_AFINET6].AfResetAddress = ResetIpv6address;
}

void
Ipv6Input(struct mbuf * bp)
{				/* IN: Pointer to IPv6 pakcet */
  u_char *cp, *ptop;
  struct mbuf *wp;
  int nb, nw;
#define IPV6IN_AFFIELD_LEN (sizeof(u_long)) /* af field passed to tun */
    /* u_char tunbuff[MAX_MRU + IPV6IN_AFFIELD_LEN]; */
  struct tun_data tun;

  tun_fill_header(tun, AF_INET6);
  cp = ptop = tun.udata.data + IPV6IN_AFFIELD_LEN;
  nb = 0;
  for (wp = bp; wp; wp = wp->next) {	/* Copy to contiguous region */
    if (sizeof tun.udata.data - (cp - tun.udata.data) < wp->cnt) {
      LogPrintf(LogERROR, "Ipv6Input: Packet too large (%d) - dropped\n",
                plength(bp));
      pfree(bp);
      return;
    }
    memcpy(cp, MBUF_CTOP(wp), wp->cnt);
    cp += wp->cnt;
    nb += wp->cnt;
  }
#ifndef NOALIAS  
  if (mode & MODE_ALIAS) {
    /* notyet; should not be reached here */
    pfree(bp);
    return;
  } else
#endif /* #ifndef NOALIAS */      
  {			/* no aliasing */
    if (PacketCheck(AF_INET6, ptop, nb, FL_IN) < 0) {
      pfree(bp);
      return;
    }
    Ipv6cpAddInOctets(nb);
    nw = write(tun_out, &tun, nb + IPV6IN_AFFIELD_LEN);
    if (nw != nb + IPV6IN_AFFIELD_LEN)
      if (nw == -1)
	LogPrintf(LogERROR, "Ipv6Input: wrote %d, got %s\n", nb,
		  strerror(errno));
      else
        LogPrintf(LogERROR, "Ipv6Input: wrote %d, got %d\n", nb, nw);
  }
  pfree(bp);

  RestartIdleTimer(AF_INET6);
#undef IPV6IN_AFFIELD_LEN
}

int
IsIpv6Enqueued()
{
  struct mqueue *queue;
  int exist = FALSE;

  for (queue = &Ipv6OutputQueues[PRI_FAST]; queue >= Ipv6OutputQueues; queue--) {
    if (queue->qlen > 0) {
      exist = TRUE;
      break;
    }
  }
  return (exist);
}

static int
SetIpv6Device(const struct sockaddr_in6 *myaddr_p,
	      struct sockaddr_in6 *hisaddr_p,
	      u_char prefixlen,
	      enum set_method how)
{
  struct sockaddr_in6 *sin6;
  int s;

  s = ID0socket(AF_INET6, SOCK_DGRAM, 0);
  if (s < 0) {
    LogPrintf(LogERROR, "SetIpv6Device: socket(): %s\n", strerror(errno));
    return (-1);
  }
  /* Init ifra */
  strncpy(ifra.ifra_name, ifrq.ifr_name,
	  IFNAMSIZ - 1); /* XXX Should be only once at somewhere. */
  
  if (how == SET_DOWN) {
    struct ppp_in6_ifaddr *pia;
    int error = 0;
    
    if (ppp_in6_ifaddr == 0) {
      close(s);
      return (0);
    }
    memset(&ifra.ifra_addr, '\0', sizeof ifra.ifra_addr);
    memset(&ifra.ifra_broadaddr, '\0', sizeof ifra.ifra_broadaddr);
    memset(&ifra.ifra_prefixmask, '\0', sizeof ifra.ifra_prefixmask);        
    pia = ppp_in6_ifaddr;
    do {
      if (ID0ioctl(s, SIOCDIFADDR_IN6, &pia->pia_areq) < 0) {
        LogPrintf(LogERROR, "SetIpv6Device: ioctl(SIOCDIFADDR_IN6): %s\n",
		  strerror(errno));
	error = -1;
      }
      pia = pia->pia_next;
    } while (pia != ppp_in6_ifaddr);
    pia = ppp_in6_ifaddr->pia_next;
    while (pia != ppp_in6_ifaddr) {
      pia = pia->pia_next;
      free(pia->pia_prev);
    }
    free(ppp_in6_ifaddr);
    ppp_in6_ifaddr = 0;
    
    close(s);
    return (error);
  } else {

    /*
     * If given addresses are already set, then ignore this request.
     */
    /* XXX If do this, need to check every address on this tun interface.
    if (IN6_ARE_ADDR_EQUAL(&old_nonlla6, myaddr_p)) {
      close(s);
      return (0);
    }
    */

    /*
     * Set interface address
     */
    sin6 = (struct sockaddr_in6 *) & (ifra.ifra_addr);
    *sin6 = *myaddr_p;

    /*
     * Set destination address if it exists.
     */
    if (hisaddr_p) {
      sin6 = (struct sockaddr_in6 *) & (ifra.ifra_dstaddr);
      *sin6 = *hisaddr_p;

      /* If hisaddr exists, set prefixlen to 0 */
      prefixlen = 0;
    }
    
    /*
     * get prefix value from prefixlen
     */
    sin6 = (struct sockaddr_in6 *)&ifra.ifra_prefixmask;
    sin6->sin6_family = 0;
    (*afsw[af_sock2ppp[AF_INET6]].AfInitSockMaskFromLen)
	((struct sockaddr *)sin6, prefixlen);
    
    if (ID0ioctl(s, SIOCAIFADDR_IN6, &ifra) < 0) {
      if (how != SET_TRY)
        LogPrintf(LogERROR, "SetIpv6Device: ioctl(SIOCAIFADDR_IN6): %s\n",
		  strerror(errno));
      close(s);
      return (-1);
    }
    else {
      /* Save this ifra to ppp_in6_ifaddr chain. */
      struct ppp_in6_ifaddr *pia;

      pia = malloc(sizeof(struct ppp_in6_ifaddr));
      if (pia == NULL) {
	LogPrintf(LogALERT,
		  "SetIpv6Device: Cannot allocate memory for a new ppp_in6_ifaddr entry\n");
      } else {
	if (ppp_in6_ifaddr) {
	  ppp_in6_ifaddr->pia_prev->pia_next = pia;
	  pia->pia_prev = ppp_in6_ifaddr->pia_prev;
	  ppp_in6_ifaddr->pia_prev = pia;
	  pia->pia_next = ppp_in6_ifaddr;
        }
	else {
	  ppp_in6_ifaddr = pia;
	  pia->pia_next = pia->pia_prev = pia;
	}
	pia->pia_areq = ifra;	    
      }
    }
  }
  close(s);
  return (0);
}

int
TrySetIpv6address(void *myaddr, void *hisaddr)
{
  return (SetIpv6Device((struct sockaddr_in6 *)myaddr,
			(struct sockaddr_in6 *)hisaddr,
			ll_prefixlen, SET_TRY));
}

int
SetIpv6address(void *myaddr, void *hisaddr)
{
  return (SetIpv6Device((struct sockaddr_in6 *)myaddr,
			(struct sockaddr_in6 *)hisaddr,
			ll_prefixlen, SET_UP));
}

int
ResetIpv6address(void)
{
  return (SetIpv6Device(0, 0, 0, SET_DOWN));
}
#endif /* INET6 */
