/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *	PPP IPV6 Control Protocol (IPV6CP) Module
 *
 *         created by copying and modifing ipcp.c
 *
 * $Id: ipv6cp.c,v 1.1.4.4 1998/07/02 14:54:17 shin Exp $
 * 
 */

/*
 *	PPP IP Control Protocol (IPCP) Module
 *
 *	    Written by Toshiharu OHNO (tony-o@iij.ad.jp)
 *
 *   Copyright (C) 1993, Internet Initiative Japan, Inc. All rights reserverd.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Internet Initiative Japan, Inc.  The name of the
 * IIJ may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: ipv6cp.c,v 1.1.4.4 1998/07/02 14:54:17 shin Exp $
 *
 *	TODO:
 *		o More RFC1772 backwoard compatibility
 */

#ifdef INET6

#include <sys/param.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>

#include <time.h>
#include <netdb.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "defs.h"
#include "command.h"
#include "hdlc.h"
#include "timer.h"
#include "fsm.h"
#include "lcpproto.h"
#include "lcp.h"
#include "multiproto.h"
#include "ipv6cp.h"
#include "os.h"
#include "phase.h"
#include "log.h"
#include "loadalias.h"
#include "vars.h"
#include "ipv6.h"
#include "throughput.h"
#include "route.h"

extern void PutConfValue();
extern void Prompt();
extern u_char * LoadIfid(u_char *ifidp, int ifid_len, int myid_flag);
extern void SaveIfid(u_char *ifidp, int myid_flag);

struct ipv6cpstate Ipv6cpInfo;
struct in6_addr Ipv6MyAddress, Ipv6HisAddress;
u_char Ipv6DefIfid[IPV6CP_IFIDLEN];

static int ipv6InOctets, ipv6OutOctets;
static int StartingIpv6In, StartingIpv6Out;

static void Ipv6cpSendConfigReq(struct fsm *);
static void Ipv6cpSendTerminateAck(struct fsm *);
static void Ipv6cpSendTerminateReq(struct fsm *);
static void Ipv6cpDecodeConfig(u_char *, int, int);
static void Ipv6cpLayerStart(struct fsm *);
static void Ipv6cpLayerFinish(struct fsm *);
static void Ipv6cpLayerUp(struct fsm *);
static void Ipv6cpLayerDown(struct fsm *);
static void Ipv6cpInitRestartCounter(struct fsm *);

static int  Ipv6cpIsUnspecIfid(u_char *cp);
static void Ipv6cpChooseNewIfid(u_char *ifidp, int myid_flag);

struct pppTimer Ipv6cpReportTimer;

struct fsm Ipv6cpFsm = {
  "IPV6CP",
  PROTO_IPV6CP,
  IPV6CP_MAXCODE,
  0,
  ST_INITIAL,
  0, 0, 0, 0,
  {0, 0, 0, NULL, NULL, NULL},	/* FSM timer */
  {0, 0, 0, NULL, NULL, NULL},	/* Open timer */
  {0, 0, 0, NULL, NULL, NULL},	/* Stopped timer */  
  LogIPV6CP,

  Ipv6cpLayerUp,
  Ipv6cpLayerDown,
  Ipv6cpLayerStart,
  Ipv6cpLayerFinish,
  Ipv6cpInitRestartCounter,
  Ipv6cpSendConfigReq,
  Ipv6cpSendTerminateReq,
  Ipv6cpSendTerminateAck,
  Ipv6cpDecodeConfig,
};

static char *cftypes[] = {
  "???", "IFID",
};

static char ifid_buf[IPV6CP_IFIDSTRLEN];
static struct pppThroughput throughput;

void
Ipv6cpAddInOctets(int n)
{
  throughput_addin(&throughput, n);
}

void
Ipv6cpAddOutOctets(int n)
{
  throughput_addout(&throughput, n);
}

static const char *
Ipv6cpPrintIfid(const u_char *src, char *dst, int size)
{
    (void)snprintf(dst, size, "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
		   src[0], src[1], src[2], src[3], src[4], src[5], src[6],
		   src[7]);
  return (dst);
}

int
ReportIpv6cpStatus(struct cmdargs const *arg)
{
  struct ipv6cpstate *icp = &Ipv6cpInfo;
  struct fsm *fp = &Ipv6cpFsm;

  if (!VarTerm)
    return 1;
  fprintf(VarTerm, "%s [%s]\n", fp->name, StateNames[fp->state]);
  if (fp->state == ST_OPENED) {  
    fprintf(VarTerm, " his side: %s\n",
	    Ipv6cpPrintIfid(icp->his_ifid, ifid_buf, sizeof(ifid_buf)));
    fprintf(VarTerm, " my  side: %s\n",
	    Ipv6cpPrintIfid(icp->want_ifid, ifid_buf, sizeof(ifid_buf)));
  }

  fprintf(VarTerm, "Defaults:\n");
  fprintf(VarTerm, " My Address:  %s\n",
	  inet_ntop(AF_INET6, &Ipv6MyAddress, ntop_buf, sizeof(ntop_buf)));
  fprintf(VarTerm, " His Address:  %s\n",
	  inet_ntop(AF_INET6, &Ipv6HisAddress, ntop_buf, sizeof(ntop_buf)));
  
  fprintf(VarTerm, "\n");
  throughput_disp(&throughput, VarTerm);  

  return 0;
}

extern int tunno;

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))

static void
rt_xaddrs(cp, cplim, rtinfo)
     caddr_t cp, cplim;
     struct rt_addrinfo *rtinfo;
{
  struct sockaddr *sa;
  int i;

  memset(rtinfo->rti_info, 0, sizeof(rtinfo->rti_info));
  for (i = 0; (i < RTAX_MAX) && (cp < cplim); i++) {
    if ((rtinfo->rti_addrs & (1 << i)) == 0)
      continue;
    rtinfo->rti_info[i] = sa = (struct sockaddr *)cp;
    cp += ROUNDUP(sa->sa_len);
  }
}

void
Ipv6cpDefIfid()
{
  int mib[6];
  size_t needed;
  char	*buf, *lim, *next, ifname[IFNAMSIZ];
  struct if_msghdr *ifm, *nextifm;
  struct ifa_msghdr *ifam = 0;
  int addrcount = 0;
  struct ifreq ifr;
  struct rt_addrinfo info;
  struct sockaddr_dl *sdl;

  memset(&Ipv6MyAddress, 0, sizeof(Ipv6MyAddress));
  memset(&Ipv6HisAddress, 0, sizeof(Ipv6HisAddress));
  memset(Ipv6DefIfid, 0, sizeof(Ipv6DefIfid));

  /* Get a link local v6addr which should have been alreadey assigned,
   * and use its Ifid value. 
   * Just do the same way as ifconfig. */
  mib[0] = CTL_NET;
  mib[1] = PF_ROUTE;
  mib[2] = 0;
  mib[3] = AF_INET6;	/* address family */
  mib[4] = NET_RT_IFLIST;
  mib[5] = 0;

  if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0) {
    LogPrintf(LogERROR,
	      "Ipv6cpDefIfid: unable to get iflist-sysctl-estimate");
    return;
  }
  if ((buf = malloc(needed)) == NULL) {
    LogPrintf(LogERROR,  "Ipv6cpDefIfid: malloc missed");
    return;
  }
  if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0) {
    LogPrintf(LogERROR,
	      "Ipv6cpDefIfid: actual retrieval of interface table failed");
    return;
  }

  ifr.ifr_addr.sa_family = AF_INET6;
  snprintf(ifname, sizeof(ifname), "tun%d", tunno);

  lim = buf + needed;    
  next = buf;
  while (next < lim) {

    ifm = (struct if_msghdr *)next;

    if (ifm->ifm_type == RTM_IFINFO) {
      sdl = (struct sockaddr_dl *)(ifm + 1);
    } else {
      LogPrintf(LogERROR,
		"Ipv6cpDefIfid: out of sync parsing NET_RT_IFLIST\n");
      LogPrintf(LogERROR, "Ipv6cpDefIfid: expected %d, got %d\n", RTM_IFINFO,
		ifm->ifm_type);
      LogPrintf(LogERROR, "Ipv6cpDefIfid: msglen = %d\n", ifm->ifm_msglen);
      LogPrintf(LogERROR, "Ipv6cpDefIfid: buf:%p, lim:%p\n", buf, lim);
      return;
    }

    next += ifm->ifm_msglen;

    ifam = NULL;
    addrcount = 0;
    while (next < lim) {
      nextifm = (struct if_msghdr *)next;
      if (nextifm->ifm_type != RTM_NEWADDR)
	break;
      if (ifam == NULL)
	ifam = (struct ifa_msghdr *)nextifm;
      addrcount++;
      next += nextifm->ifm_msglen;
    }
    if (ifam == NULL)
      continue;
    if (next >= lim)
	break;
    if (strlen(ifname) != sdl->sdl_nlen)
      continue; /* not same len */

    if (strncmp(ifname, sdl->sdl_data, sdl->sdl_nlen) != 0)
      continue; /* not same name */

    LogPrintf(LogIPV6CP,
	      "Ipv6cpDefIfid: ifname matched, name %s, len %d\n", ifname,
	      sdl->sdl_nlen);

    while (addrcount > 0) {
      info.rti_addrs = ifam->ifam_addrs;
      /* Expand the compacted addresses */
      rt_xaddrs((char *)(ifam + 1), ifam->ifam_msglen + (char *)ifam,
		&info);
    
      if (info.rti_info[RTAX_IFA]->sa_family == AF_INET6 &&
	IN6_IS_ADDR_LINKLOCAL
	  (&((struct sockaddr_in6 *)info.rti_info[RTAX_IFA])->sin6_addr)) {
	  Ipv6MyAddress =
	      ((struct sockaddr_in6 *)info.rti_info[RTAX_IFA])->sin6_addr;
	  break;
      }
      addrcount--;
      ifam = (struct ifa_msghdr *)((char *)ifam + ifam->ifam_msglen);
    }
  }
  /* save default Interface ID */
  memcpy(Ipv6DefIfid, IPV6CP_IFIDP(Ipv6MyAddress), IPV6CP_IFIDLEN);
}

void
Ipv6cpInit()
{
  struct ipv6cpstate *icp = &Ipv6cpInfo;

  FsmInit(&Ipv6cpFsm);
  memset(icp, '\0', sizeof(*icp));
  if ((mode & MODE_DEDICATED) && !GetLabel()) {
    memset(icp->want_ifid, '\0', IPV6CP_IFIDLEN);
    memset(icp->his_ifid, '\0', IPV6CP_IFIDLEN);
  } else {
    memcpy(icp->want_ifid, IPV6CP_IFIDP(Ipv6MyAddress), IPV6CP_IFIDLEN);
    memcpy(icp->his_ifid, IPV6CP_IFIDP(Ipv6HisAddress), IPV6CP_IFIDLEN);    
  }

  Ipv6cpFsm.maxconfig = 10;
  throughput_init(&throughput);  

  /* afsw initialization */
  afsw[PPP_AFINET6].AfMyAddr = &Ipv6MyAddress;
  afsw[PPP_AFINET6].AfHisAddr = &Ipv6HisAddress;
}

static void
Ipv6cpInitRestartCounter(struct fsm * fp)
{
  fp->FsmTimer.load = VarRetryTimeout * SECTICKS;
  fp->restart = 5;
}

static void
Ipv6cpSendConfigReq(struct fsm * fp)
{
  u_char *cp;
  struct ipv6cpstate *icp = &Ipv6cpInfo;
  struct lcp_opt o;

  cp = ReqBuff;
  LogPrintf(LogIPV6CP, "Ipv6cpSendConfigReq\n");
  if (
#ifndef CONNECT_WITH_SPPP
      !DEV_IS_SYNC ||
#endif /* !CONNECT_TO_SPPP */
      !REJECTED(icp, TY_IFID)) {
    o.id = TY_IFID;
    o.len = 10;
    if ((mode & (MODE_DIRECT|MODE_DEDICATED)) &&
	Ipv6cpIsUnspecIfid(icp->want_ifid))
      Ipv6cpChooseNewIfid(icp->want_ifid, 1);
    memcpy(o.data, icp->want_ifid, IPV6CP_IFIDLEN);
    cp += LcpPutConf(LogIPV6CP, cp, &o, cftypes[o.id],
		     Ipv6cpPrintIfid(icp->want_ifid, ifid_buf,
				     sizeof(ifid_buf)));
  }

  FsmOutput(fp, CODE_CONFIGREQ, fp->reqid++, ReqBuff, cp - ReqBuff);
}

static void
Ipv6cpSendTerminateReq(struct fsm * fp)
{
  LogPrintf(LogIPV6CP, "Ipv6cpSendTerminateReq\n");
}

static void
Ipv6cpSendTerminateAck(struct fsm * fp)
{
  LogPrintf(LogIPV6CP, "Ipv6cpSendTerminateAck\n");
  FsmOutput(fp, CODE_TERMACK, fp->repid, NULL, 0);
}

static void
Ipv6cpLayerStart(struct fsm * fp)
{
  LogPrintf(LogIPV6CP, "Ipv6cpLayerStart.\n");
}

static void
Ipv6cpLayerFinish(struct fsm * fp)
{
  AfActiveFlag &= ~AFACTIVE_INET6;
  LogPrintf(LogIPV6CP, "Ipv6cpLayerFinish.\n");
  Ipv6cpSendTerminateReq(&Ipv6cpFsm);
  if (AfActiveFlag == 0) {
    reconnect(RECON_FALSE);
    LcpClose();
    NewPhase(PHASE_TERMINATE);
  }
}

int
Ipv6cpOctetsIn()
{
  return ipv6InOctets < StartingIpv6In ?
    INT_MAX - StartingIpv6In + ipv6InOctets - INT_MIN + 1 :
    ipv6InOctets - StartingIpv6In;
}

int
Ipv6cpOctetsOut()
{
  return ipv6OutOctets < StartingIpv6Out ?
    INT_MAX - StartingIpv6Out + ipv6OutOctets - INT_MIN + 1 :
    ipv6OutOctets - StartingIpv6Out;
}

static void
Ipv6cpLayerDown(struct fsm * fp)
{
  LogPrintf(LogIPV6CP, "Ipv6cpLayerDown.\n");
  throughput_stop(&throughput);
  throughput_log(&throughput, LogIPV6CP, NULL);
  ResetIpv6address();  
}

/*
 *  Called when IPV6CP has reached to OPEN state
 */
static void
Ipv6cpLayerUp(struct fsm * fp)
{
  char tbuff[100];
  struct sockaddr_in6 sin6;

  memset(&sin6, 0, sizeof(sin6));
  sin6.sin6_len = sizeof(sin6);
  sin6.sin6_family = AF_INET6;

  AfActiveFlag |= AFACTIVE_INET6;
  Prompt();
  LogPrintf(LogIPV6CP, "Ipv6cpLayerUp(%d).\n", fp->state);
  snprintf(tbuff, sizeof(tbuff), "my interface id = %s ",
	   Ipv6cpPrintIfid(Ipv6cpInfo.want_ifid, ifid_buf, sizeof(ifid_buf)));
  
  LogPrintf(LogIsKept(LogIPV6CP) ? LogIPV6CP : LogLINK,
	    " %s his interface id = %s\n", tbuff, 
	    Ipv6cpPrintIfid(Ipv6cpInfo.his_ifid, ifid_buf, sizeof(ifid_buf)));
  /* Init myaddr and hisaddr */
  if (memcmp(&Ipv6MyAddress, &in6addr_any,
	     sizeof(in6addr_any) - IPV6CP_IFIDLEN) == 0) {
    char ifname[IFNAMSIZ];

    snprintf(ifname, sizeof(ifname), "tun%d", tunno);
    Ipv6MyAddress.s6_addr16[0] = IPV6_ADDR_INT16_ULL;
#ifdef SIN6_IFINDEX
    sin6.sin6_ifindex = if_nametoindex(ifname);
#else /* SIN6_IFINDEX */
    Ipv6MyAddress.s6_addr16[1] = htons((u_int16_t)if_nametoindex(ifname));
#endif /* SIN6_IFINDEX */
  }
  memcpy(&Ipv6HisAddress, &Ipv6MyAddress,
	 sizeof(Ipv6HisAddress) - IPV6CP_IFIDLEN);
  memcpy(IPV6CP_IFIDP(Ipv6MyAddress), Ipv6cpInfo.want_ifid, IPV6CP_IFIDLEN);
  memcpy(IPV6CP_IFIDP(Ipv6HisAddress), Ipv6cpInfo.his_ifid, IPV6CP_IFIDLEN);
  sin6.sin6_addr = Ipv6MyAddress;
  if (memcmp(IPV6CP_IFIDP(Ipv6MyAddress), Ipv6DefIfid, IPV6CP_IFIDLEN)) {
    if (SetIpv6address(&sin6, 0) < 0) {
      if (VarTerm)
        LogPrintf(LogERROR, "Ipv6cpLayerUp: unable to set ipv6 address\n");
      return;
    }
  }
  { /* XXX */
    struct sockaddr_in6 dest6, gateway6, netmask6;
    memset(&netmask6, 0, sizeof(netmask6));
    dest6 = gateway6 = sin6;
    netmask6.sin6_len = sizeof(netmask6);
    netmask6.sin6_family = 0;

    (*afsw[PPP_AFINET6].AfInitSockMaskFromLen)((struct sockaddr *)&netmask6,
					       ll_prefixlen);
    OsSetRoute(RTM_ADD, (struct sockaddr *)&dest6,
	       (struct sockaddr *)&gateway6,
	       (struct sockaddr *)&netmask6,
	       RTF_UP|RTF_STATIC /* to deny RTF_GATEWAY */,
	       1 /* force change when duped */);
  }
  throughput_start(&throughput);  
  ProtoLinkup(AF_INET6);
  StartIdleTimer();
}

void
Ipv6cpUp()
{
  FsmUp(&Ipv6cpFsm);
  LogPrintf(LogIPV6CP, "IPV6CP Up event!!\n");
}

void
Ipv6cpOpen()
{
  FsmOpen(&Ipv6cpFsm);
}

static int
Ipv6cpIsUnspecIfid(u_char *cp)
{
  int i;

  for (i = 0; i < IPV6CP_IFIDLEN; i++) {
    if (cp[i])
      return (0);
  }
  return (1);
}

static int
AcceptableIfid(char *rcvd_ifid, char * tgt_ifid)
{
  LogPrintf(LogDEBUG, "requested interface id = %x\n",
	    Ipv6cpPrintIfid(rcvd_ifid, ifid_buf, sizeof(ifid_buf)));
  LogPrintf(LogDEBUG, "tgt interface id = %x\n",
	    Ipv6cpPrintIfid(tgt_ifid, ifid_buf, sizeof(ifid_buf)));

  return (!Ipv6cpIsUnspecIfid(rcvd_ifid) &&
	  BCMP(rcvd_ifid, tgt_ifid, IPV6CP_IFIDLEN));
}

/*
 * Just use time val as new Ifid, and clear the global bit.
 */
static void
Ipv6cpChooseNewIfid(u_char *ifidp, int myid_flag)
{
  struct timeval ct;
  u_char saved_ifid[INET6_ADDRSTRLEN];
  struct in6_addr ifid_wnp; /* ifid With Null Prefix */
  u_char *saved_ifidp;

  saved_ifid[0] = '\0';
  saved_ifidp = LoadIfid(saved_ifid, sizeof(saved_ifid), myid_flag);
  if (inet_pton(AF_INET6, saved_ifidp, &ifid_wnp) == 0)
    ifid_wnp = in6addr_any;
  if (Ipv6cpIsUnspecIfid(IPV6CP_IFIDP(ifid_wnp)) ||
      memcmp(ifidp, IPV6CP_IFIDP(ifid_wnp), IPV6CP_IFIDLEN) == 0) {
    gettimeofday(&ct, NULL);
    /* XXX; IPV6CP_IFIDLEN is assumed to be 8 below. */
    memcpy(IPV6CP_IFIDP(ifid_wnp), (char *)&ct.tv_sec, 4);
    memcpy(IPV6CP_IFIDP(ifid_wnp) + 4, (char *)&ct.tv_usec, 4);
    (IPV6CP_IFIDP(ifid_wnp))[0] &= ~0x2; /* clear the global bit */
  }
  memcpy(ifidp, IPV6CP_IFIDP(ifid_wnp), IPV6CP_IFIDLEN);
  inet_ntop(AF_INET6, &ifid_wnp, saved_ifid, sizeof(saved_ifid));
  SaveIfid(saved_ifid, myid_flag);
}

static void
Ipv6cpDecodeConfig(u_char *cp, int plen, int mode)
{
  int type, length;
  char tbuff[100];
  char tbuff2[100];
  char *opt_info_p;

  ackp = AckBuff;
  nakp = NakBuff;
  rejp = RejBuff;

  while (plen >= sizeof(struct fsmconfig)) {
    type = *cp;
    length = cp[1];
    if (type <= TY_IFID)
      snprintf(tbuff, sizeof(tbuff), " %s[%d] ", cftypes[type], length);
    else
      snprintf(tbuff, sizeof(tbuff), " ");

    switch (type) {
    case TY_IFID:		/* draft-ietf-ipngwg-ipv6-over-ppp-05.txt */
      opt_info_p = cp + 2;
      LogPrintf(LogIPV6CP, "%s %s\n", tbuff,
		Ipv6cpPrintIfid(opt_info_p, ifid_buf, sizeof(ifid_buf)));
      switch (mode) {
      case MODE_REQ:
	if (!AcceptableIfid(opt_info_p, Ipv6cpInfo.want_ifid)) {
	  /* If my ifid is 0, it means either ifid is 0, then send Conf Rej. */
	  if (Ipv6cpIsUnspecIfid(Ipv6cpInfo.want_ifid)) {
	    memcpy(rejp, cp, 2);
	    memcpy(rejp + 2, opt_info_p, length - 2);
	    rejp += length;
	    break;
	  }
	  /*
	   * Else if His ifid is 0 or same as us, choose new Ifid for him.
	   */
	  Ipv6cpChooseNewIfid(Ipv6cpInfo.his_ifid, 0);
	  memcpy(nakp, cp, 2);
	  memcpy(nakp + 2, Ipv6cpInfo.his_ifid, length);
	  nakp += length;
	  break;
	}
	memcpy(Ipv6cpInfo.his_ifid, opt_info_p, IPV6CP_IFIDLEN);
	memcpy(ackp, cp, length);
	ackp += length;
	break;
      case MODE_NAK:
	if (AcceptableIfid(opt_info_p, Ipv6cpInfo.his_ifid)) {
	  /*
	   * Use address suggested by peer.
	   */
	  snprintf(tbuff2, sizeof(tbuff2), "%s changing interface id: %s ",
		   tbuff,
		   Ipv6cpPrintIfid(Ipv6cpInfo.want_ifid, ifid_buf,
				   sizeof(ifid_buf)));
	  LogPrintf(LogIPV6CP, "%s --> %s\n", tbuff2,
		    Ipv6cpPrintIfid(opt_info_p, ifid_buf, sizeof(ifid_buf)));
	  memcpy(Ipv6cpInfo.want_ifid, opt_info_p, IPV6CP_IFIDLEN);
	  /* Save new my ifid to a file */
	  {
	    struct in6_addr ifid_wnp; /* interface id With Null Prefix */
	    u_char saved_ifid[INET6_ADDRSTRLEN];

	    memset(&ifid_wnp, 0, sizeof(ifid_wnp));
	    memcpy(IPV6CP_IFIDP(ifid_wnp), opt_info_p, IPV6CP_IFIDLEN);
	    inet_ntop(AF_INET6, &ifid_wnp, saved_ifid, sizeof(saved_ifid));
	    SaveIfid(saved_ifid, 1);
	  }
	}
	else if (!Ipv6cpIsUnspecIfid(opt_info_p)) {
	  /*
	   * His ifid is non 0 and same as us, choose new Ifid for me.
	   */
	  Ipv6cpChooseNewIfid(Ipv6cpInfo.want_ifid, 1);
	  Ipv6cpInitRestartCounter(&Ipv6cpFsm);
	} /* else ignore it */
	break;
      case MODE_REJ:
	Ipv6cpInfo.his_reject |= (1 << type);
	break;
      }
      break;
    default:
      Ipv6cpInfo.my_reject |= (1 << type);
      memcpy(rejp, cp, length);
      rejp += length;
      break;
    }
    plen -= length;
    cp += length;
  }
}

void
Ipv6cpInput(struct mbuf * bp)
{
  FsmInput(&Ipv6cpFsm, bp);
}

#endif /* INET6 */
