/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *		PPP Header for Multiprotocol common procedures
 *
 * $Id: multiproto.h,v 1.1.4.1 1998/04/20 13:03:26 shin Exp $
 *
 */
#ifndef _MULTIPROTO_H_
#define	_MULTIPROTO_H_

struct _sockaddr_max {
  u_char sa_len;
  u_char sa_family;
  char   sa_data[
#ifdef INET6
		 22
#else /* INET6 */
		 14
#endif /* INET6 */
		 ];
};

#define PPP_AFUNSPEC	0
#define PPP_AFINET	(PPP_AFUNSPEC + 1)
#define PPP_AFMIN	PPP_AFINET
#define PPP_AFINET6	(PPP_AFINET + 1)
#define PPP_AFMAX	(PPP_AFINET6 + 1)

#ifdef INET
#define DEF_SOCK_AF AF_INET
#elif INET6
#define DEF_SOCK_AF AF_INET6
#endif /* INET or INET6 */

struct afsw {
  char AfAddrOffInSock;
  char *AfMyAddrLabel;
  int AfFlags;
  int *AfIdleTime;
  int AfLogLevel;
  int AfMbInType;
  int AfMbQType;
  struct fsm *AfFsm;
  struct mqueue *AfOutputQueues;
  void *AfMyAddr;
  void *AfHisAddr;
  void (*AfErrorReport) (void * pr_hdr_p, int code);
  void (*AfInitSockMaskFromLen) (struct sockaddr *sa, int bitlen);
  void (*AfInput) (struct mbuf *bp);
  void (*AfMakeAddrUnspec) (struct sockaddr *sa);
  void (*AfStartOutput) (void);
  void (*AfSendPppFrame) (struct mbuf *bp);
  u_short (*AfPacketLen) (void * pr_hdr_p);
  int (*AfIsAddrUnspec) (struct sockaddr *sa);
  int (*AfLoopbackCheck) (void *pr_hdr_p);
  int (*AfSetAddress) (void *, void *);
  int (*AfResetAddress) (void);
};
#define PPP_AFFLAGS_HAS_ALIAS	0x1
#define PPP_AFFLAGS_LINKUP	0x2
#define PPP_AFFLAGS_SUPPORTED	0x4

enum set_method { SET_UP, SET_DOWN, SET_TRY };

extern struct afsw afsw[PPP_AFMAX];
extern int af_sock2ppp[AF_MAX];
extern int af_ppp2sock[PPP_AFMAX];
extern char ntop_buf[
#ifdef INET6
		     INET6_ADDRSTRLEN
#elif INET
		     INET_ADDRSTRLEN
#endif /* INET6, INET */
		     ];
#define _INET_NTOP(af, sa) \
	inet_ntop(af, \
		  (char *)sa + afsw[af_sock2ppp[af]].AfAddrOffInSock, \
		  ntop_buf, \
		  sizeof(ntop_buf))
extern int AfActiveFlag;

#define AFACTIVE_INET	0x1
#define AFACTIVE_INET6	0x2

int PacketCheck(u_long af, void *cp, int nb, int direction);

void AfEnqueue(struct afsw *afsw_p, int pri, char *ptr, int count);
void AfInit(void);

void IdleTimeout(void *v);
void StartIdleTimer(void);
void UpdateIdleTimer(void);
void StopIdleTimer(void);
void RestartIdleTimer(u_long af);
int  RemainingIdleTime(void);

int  AfLinkdown(void);
int  ProtoInterfaceDown(int);
void ProtoLinkup(u_long af);
void PppLinkdown(void);
void ProtoInitLenFromSockMask(u_long af, struct sockaddr *sa);
void ProtoStartOutput(void);
#endif /* _MULTIPROTO_H_ */
